// Copyright 1995 Barbara Liskov


#include "parse.h"
#include "cg.h"
#include "types/str.h"

extern int version;

//	This file handles inlining of a limited number of for statements.
//	Currently these invoke the
//		the "to" iterator of "int"
//		the "elements" and "indexes" iterators of vector
//		the "indexes" iterator of "array"
//
//	It seems that "to_by" of "int" is a little tricky in terms
//		of suppressing wrap around.
//
//	It seems that "elements" of "array" is huge, in terms of code space.
//

static string int_str = NULL;
static string bool_str = NULL;
extern void handle_body(Body *body);
extern void handle_body_1(Body *body);
extern bool emit_iter_prefix();
Expr *opt_inv_get_prim(Invoc *inv, Id **id);
void expr_emit_local(Expr *ex);
bool opt_for(iter_type which, Id *id, Invoc *inv, Body* body);
static bool opt_for0(iter_type which, Id *id, Invoc *inv, Body* body);
static bool opt_for1(iter_type which, Id *id, Invoc *inv, Body* body);


//	Do decls for temps necessary for inlined iterators

void opt_inv_temps(Invoc *inv)
{
iter_type which;

	if (!int_str) {
		int_str = string_new("int");
		bool_str = string_new("bool");
		}
	bool ok = opt_inv(inv, &which);
	if (!ok) return;
	switch (which) {
		case INT_TO: {
        		if (version == 2 && get_iter_active()) return;
			save_temp(int_str);
			return;
			}
		case INT_TO_BY: {
        		if (version == 2 && get_iter_active()) return;
			// save_temp(int_str);
			// save_temp(int_str);
			return;
			}
		case ARRAY_ELTS: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct arrayMelements_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// if (version == 2) {
				// save_temp(int_str);
			 	// save_temp(int_str);
				// }
			return;
			}
		case ARRAY_INDX: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct arrayMindexes_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// save_temp(int_str);
			// if (version == 2) save_temp(int_str);
			return;
			}
		case VECTOR_ELTS: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct vectorMelements_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// save_temp(int_str);
			// save_temp(int_str);
			return;
			}
		case VECTOR_INDX: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct vectorMindexes_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// save_temp(int_str);
			return;
			}
		case SEQUENCE_ELTS: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct sequenceMelements_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// save_temp(int_str);
			// save_temp(int_str);
			return;
			}
		case SEQUENCE_INDX: {
        		if (version == 2 && get_iter_active()) {
				string nm = string_new(
				   "struct sequenceMindexes_macro_Iter_State_s");
                                save_temp(nm);
				save_temp(bool_str);
				return;
				}
			// save_temp(int_str);
			return;
			}
		}
	}

// emit code for inlined iterator for declforstmt

bool opt_declforstmt(DeclForStmt *declforstmt)
{
Invoc *inv = declforstmt->get_invoc();  
iter_type which;
ParseNodeList *ids = NULL;

	// Is this an inlineable iterator?
	// (Need to decide if it's inlineable *before* any code is emitted...)

	bool ok = opt_inv(inv, &which);
	if (!ok) return FALSE;

	// fish out the decl and its id

	ParseNodeList *decls = declforstmt->get_decls();
	if (!decls || decls->length() != 1) {
			cmp_err("Inlining is incorrect.",
				inv->get_line());
			return FALSE;
			}
	Decl *d = (Decl *)(decls->front());
   	switch (d->tag()) {
       		 case Decl::ImplDeclT: {
//		 case Decl::VarArgsDeclT: {
       		         cmp_err("Improper decl in declforstmt.", 
       		                 declforstmt->get_line());
       		         return FALSE;
       		         }
       		 case Decl::RegDeclT: {
       		         RegDecl *rd = (RegDecl *)(d);
       		         ids = rd->get_ids();
       		         }
       		 }

	if (!ids || ids->length() != 1) {
			cmp_err("Inlining is incorrect.",
				inv->get_line());
			return FALSE;
			}
	Id *id = (Id *)(ids->front());

	Body *body = declforstmt->get_body();

	return opt_for(which, id, inv, body);
	}

bool opt_for(iter_type which, Id *id, Invoc *inv, Body* body)
{
	if (version == 0 || version == 1) {
		return opt_for0(which, id, inv, body);
		}
	else {
		return opt_for1(which, id, inv, body);
		}
	}

bool opt_for0(iter_type which, Id *id, Invoc *inv, Body* body)
{
Id *mid;
ParseNodeList *exprs;

	// do preprocessing
	ParseNode *new_pn = inv_preprocess(inv);
	if (new_pn->tag() != ParseNode::InvocT) {
			cmp_err("Inlining is incorrect.",
				inv->get_line());
			return FALSE;
			}
	Invoc *new_inv = (Invoc *)new_pn;

	// now get the primary
	Expr *prim = opt_inv_get_prim(new_inv, &mid);
	exprs = new_inv->get_exprs();


	// do the code

	string nm = id->get_id();
	switch (which) {
		case INT_TO: {
   			ind();
			string dest_str = get_next_temp();
			emit_iter_prefix();
			pstr(dest_str);
			ps(" = ");
			expr_emit((Expr *)(exprs->front()));
			ps(";\n");
   			ind();

   			ps("for (");
			pstr(nm);
			ps(" = ");
			expr_emit(prim);
			ps("; ");
			pstr(nm);
			ps(" <= ");
			emit_iter_prefix();
			pstr(dest_str);
			ps("; ");
			pstr(nm);
			ps("++)");

			ps(" {\n");
			handle_body(body);
			ind();
			ps("}\n");
			return TRUE;
			}

		case ARRAY_INDX: {
			ind();
			ps("FIX(");
			expr_emit(prim);
			ps(", array);\n");
   			ind();
			ps("FIXUPREAD(&((array_C)");
			expr_emit(prim);
			ps(")->hdr.inh);\n");
   			ind();
			string limit_str = get_next_temp();
			emit_iter_prefix();
			pstr(limit_str);
			ps(" = ((array_C)");
			expr_emit(prim);
			ps(")->user_low + ((array_C)");
			expr_emit(prim);
			ps(")->user_size - 1");
			ps(";\n");
   			ind();

   			ps("for (");
			pstr(nm);
			ps(" = ((array_C)");
			expr_emit(prim);
			ps(")->user_low; ");
			pstr(nm);
			ps(" <= ");
			emit_iter_prefix();
			pstr(limit_str);
			ps("; ");
			pstr(nm);
			ps("++)");

			ps(" {\n");
			handle_body(body);
			ind();
			ps("}\n");
			return TRUE;
			}

		case VECTOR_ELTS: {
			TypeInterface *ti = inv->get_type();
			vec v = ti->get_mult();
			if (!v || vec_length(v) != 1) {
				cmp_err("Optimization failed", inv->get_line());
				return FALSE;
				}
			type t = UNPV(type, vec_fetch(v, 0));
			string type_str = simple_type_name(type_name(t));

			ind();
			ps("FIX(");
			expr_emit(prim);
			ps(", vector);\n");
   			ind();
			ps("FIXUPREAD(&((vector_C)");
			expr_emit(prim);
			ps(")->hdr.inh);\n");
   			ind();
			string index_str = get_next_temp();
			string limit_str = get_next_temp();
			emit_iter_prefix();
			pstr(limit_str);
			ps(" = ((vector_C)");
			expr_emit(prim);
			ps(")->hdr.inh.size;\n");
   			ind();

   			ps("for (");
			emit_iter_prefix();
			pstr(index_str);
			ps(" = 0; ");
			emit_iter_prefix();
			pstr(index_str);
			ps(" < ");
			emit_iter_prefix();
			pstr(limit_str);
			ps("; ");
			emit_iter_prefix();
			pstr(index_str);
			ps("++)");


			ps(" {\n");
			ind();
			ps("DISCOVER(((vector_C)");
			expr_emit(prim);
			ps(")->items[");
			emit_iter_prefix();
			pstr(index_str);
			ps("], pval);\n");
			ind();
			pstr(nm);
			ps(" = UNPV(");
			pstr(type_str);
			ps(", ((vector_C)");
			expr_emit(prim);
			ps(")->items[");
			emit_iter_prefix();
			pstr(index_str);
			ps("]);\n");
			
			handle_body(body);
			ind();
			ps("}\n");
			return TRUE;
			}

		case VECTOR_INDX: {
			ind();
			ps("FIX(");
			expr_emit(prim);
			ps(", vector);\n");
   			ind();
			ps("FIXUPREAD(&((vector_C)");
			expr_emit(prim);
			ps(")->hdr.inh);\n");
   			ind();
			string limit_str = get_next_temp();
			emit_iter_prefix();
			pstr(limit_str);
			ps(" = ((vector_C)");
			expr_emit(prim);
			ps(")->hdr.inh.size;\n");
   			ind();

   			ps("for (");
			pstr(nm);
			ps(" = 1; ");
			pstr(nm);
			ps(" <= ");
			emit_iter_prefix();
			pstr(limit_str);
			ps("; ");
			pstr(nm);
			ps("++)");

			ps(" {\n");
			handle_body(body);
			ind();
			ps("}\n");
			return TRUE;
			}
		}
	return FALSE;
	}

bool opt_for1(iter_type which, Id *id, Invoc *inv, Body* body)
{
Id *mid;
ParseNodeList *exprs;

	// do preprocessing
	ParseNode *new_pn = inv_preprocess(inv);
	if (new_pn->tag() != ParseNode::InvocT) {
			cmp_err("Inlining is incorrect.",
				inv->get_line());
			return FALSE;
			}
	Invoc *new_inv = (Invoc *)new_pn;

	// now get the primary
	Expr *prim = opt_inv_get_prim(new_inv, &mid);
	exprs = new_inv->get_exprs();


	// do the code

	string nm = id->get_id();
	type t = get_one_type(id->get_type(), 300);
	switch (which) {
		case INT_TO: {
			loop_enter();
   			ind();
			string dest_str = get_next_temp();
			emit_iter_prefix();
			pstr(dest_str);
			ps(" = ");
			expr_emit((Expr *)(exprs->front()));
			ps(";\n");
   			ind();

   			ps("for (");
			emit_iter_prefix();
			pstr(nm);
			ps(" = ");
			expr_emit(prim);
			ps("; ");
			emit_iter_prefix();
			pstr(nm);
			ps(" <= ");
			emit_iter_prefix();
			pstr(dest_str);
			ps("; ");
			emit_iter_prefix();
			pstr(nm);
			ps("++)");

			ps(" {\n");
			handle_body_1(body);
			ind();
			ps("}\n");
			loop_end_label();
			loop_exit();
			return TRUE;
			}

//		case ARRAY_INDX: {
//			ind();
//			ps("{\n");
//			ind();
//			ps("struct Array_f_s *");
//			expr_emit(prim);
//			ps("FS = ");
//			// ind();
//			// ps("SETUP_SELF(");
//			// expr_emit(prim);
//			// ps(", Array);\n");
//			ps("RESIDENT_CORE(");
//			expr_emit(prim);
//			ps(", Array);\n");
//			ind();
//			ps("PREPARE_READ_F(");
//			expr_emit(prim);
//			ps(", ");
//			expr_emit(prim);
//			ps("FS);\n");
//			ind();
//			ps("MARK_USED_F(");
//			expr_emit(prim);
//			ps("FS);\n");
//   			// ind();
//			// ps("FIXFIELDS(");
//			// expr_emit(prim);
//			// ps(", Array);\n");
//   			// ind();
//			// string low_str = get_next_temp();
//			// ps("GETV_F(");
//			// pstr(low_str);
//			// ps(", ");
//			// expr_emit(prim);
//			// ps("FS, user_low, Array);\n");
//
//   			ind();
//			string limit_str = get_next_temp();
//			ps("GETV_F(");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps(", ");
//			expr_emit(prim);
//			ps("FS, user_size, Array);\n");
//
//			// ind();
//			// emit_iter_prefix();
//			// pstr(limit_str);
//			// ps(" = ");
//			// pstr(low_str);
//			// ps(" + ");
//			// emit_iter_prefix();
//			// pstr(limit_str);
//			// ps(" - 1;\n");
//
//   			ind();
//   			ps("for (");
//			emit_iter_prefix();
//			pstr(nm);
//			ps(" = 0; ");
//			emit_iter_prefix();
//			pstr(nm);
//			ps(" < ");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps("; ");
//			emit_iter_prefix();
//			pstr(nm);
//			ps("++)");
//
//			ps(" {\n");
//			handle_body_1(body);
//			ind();
//			ps("}\n");
//			ps("}\n");
//			return TRUE;
//			}
//
		case ARRAY_INDX: {
		   loop_enter();
		   ind();
		   ps("{\n");
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			ind();
			ps("ARRAY_GEN_INDEXES_START(");
			expr_emit(prim);
			ps(", (");
                        emit_iter_prefix();
			pstr(state);
			ps("));\n");
			ind();
			ps("while (1) {\n");
			ind();
			ps("ARRAY_GEN_INDEXES_LOOP((");
                        emit_iter_prefix();
			pstr(state);
			ps("), ");
                        emit_iter_prefix();
			pstr(nm);
			ps(", ");
                        emit_iter_prefix();
			pstr(cont);
			ps(");\n");
			ind();
			ps("if (!");
			emit_iter_prefix();
			pstr(cont);
			ps(") break;\n");
			handle_body_1(body);
			ind();
			ps("}\n");
			}
		   else {
			ind();
			ps("ARRAY_INDEXES_START(");
			expr_emit(prim);
			ps(", ");
			emit_iter_prefix();
			pstr(nm);
			ps(");\n");
			handle_body_1(body);
			ind();
			ps("ARRAY_INDEXES_END();\n");
			}
		   ind();
		   ps("}\n");
		   loop_end_label();
		   loop_exit();
		   return TRUE;
		   }

		case ARRAY_ELTS: {
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			loop_enter();
			ind();
			ps("{\n");
			ind();
			if (is_vtype(t)) {
				ps("ARRAY_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("ARRAY_GEN_PRIMITIVE_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), ");
                        	emit_iter_prefix();
				pstr(nm);
				ps(", ");
				pstr(simple_type_name(type_name(t)));
				ps(", (");
                        	emit_iter_prefix();
				pstr(cont);
				ps("));\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
				}
			else {
				ps("ARRAY_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("ARRAY_GEN_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), MT_1, ");
                        	emit_iter_prefix();
				pstr(cont);
				ps(");\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
                        	ind();
                        	emit_iter_prefix();
                        	pstr(nm);
                        	ps(" = PV_as_obj");
                        	ps("(");
                        	pstr(simple_type_name(type_name(t)));
                        	ps(", MT_1);\n");
				}
			handle_body_1(body);
			ind();
			ps("}\n");
			ind();
			ps("}\n");
			loop_end_label();
			loop_exit();
			return TRUE;
			}
		   else {
			loop_enter();
			ind();
			ps("{\n");
			ind();
			if (is_vtype(t)) {
				ps("ARRAY_PRIMITIVE_ELEMENTS_START(");
				expr_emit(prim);
				ps(", ");
                        	emit_iter_prefix();
				pstr(nm);
				ps(", ");
				pstr(simple_type_name(type_name(t)));
				ps(");\n");
				}
			else {
				ps("ARRAY_ELEMENTS_START(");
				expr_emit(prim);
                        	ps(", MT_1);\n");
                        	ind();
                        	emit_iter_prefix();
                        	pstr(nm);
                        	ps(" = PV_as_obj");
                        	ps("(");
                        	pstr(simple_type_name(type_name(t)));
                        	ps(", MT_1);\n");
				}
			handle_body_1(body);
			ind();
			ps("ARRAY_ELEMENTS_END();\n");
			ind();
			ps("}\n");
			loop_end_label();
			loop_exit();
			return TRUE;
			}
		   }

//		case ARRAY_ELTS: {
//			ind();
//			ps("{\n");
//			ind();
//			ps("struct Array_f_s *");
//			expr_emit(prim);
//			ps("FS = ");
//			ps("RESIDENT_CORE(");
//			expr_emit(prim);
//			ps(", Array);\n");
//			ind();
//			ps("PREPARE_READ_F(");
//			expr_emit(prim);
//			ps(", ");
//			expr_emit(prim);
//			ps("FS);\n");
//			ind();
//			ps("MARK_USED_F(");
//			expr_emit(prim);
//			ps("FS);\n");
//
//   			ind();
//			string limit_str = get_next_temp();
//			ps("GETV_F(");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps(", ");
//			expr_emit(prim);
//			ps("FS, user_size, Array);\n");
//
//   			ind();
//			string index_str = get_next_temp();
//   			ps("for (");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps(" = 0; ");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps(" < ");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps("; ");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps("++)");
//
//			ps(" {\n");
//			ind();
//			ps("array_inline_fetch(MT_1, ");
//			emit_iter_prefix();
//			expr_emit(prim);
//			ps(", ");
//			emit_iter_prefix();
//			expr_emit(prim);
//			ps("FS, ");
//			pstr(index_str);
//			ps(");\n");
//			ind();
//			emit_iter_prefix();
//			pstr(nm);
//			ps(" = PV_as_");
//			if (is_vtype(t)) ps("prim");
//			else ps("obj");
//			ps("(");
//			pstr(simple_type_name(type_name(t)));
//			ps(", MT_1);\n");
//			handle_body_1(body);
//			ind();
//			ps("}\n");
//			ps("}\n");
//			return TRUE;
//			}

//		case VECTOR_ELTS: {
//                        loop_enter();
//                        ind();
//                        ps("{\n");
//                        ind();
//                        ps("VECTOR_ELEMENTS_START(");
//                        expr_emit(prim);
//                        ps(", MT_1);\n");
//                        ind();
//                        emit_iter_prefix();
//                        pstr(nm);
//                        ps(" = PV_as_");
//                        if (is_vtype(t)) ps("prim");
//                        else ps("obj");
//                        ps("(");
//                        pstr(simple_type_name(type_name(t)));
//                        ps(", MT_1);\n");
//                        handle_body_1(body);
//                        ind();
//                        ps("VECTOR_ELEMENTS_END();\n");
//                        ind();
//                        ps("}\n");
//                        loop_end_label();
//                        loop_exit();
//                        return TRUE;
//                        }

//		case VECTOR_ELTS: {
//			loop_enter();
//			TypeInterface *ti = inv->get_type();
//			vec v = ti->get_mult();
//			if (!v || vec_length(v) != 1) {
//				cmp_err("Optimization failed", inv->get_line());
//				return FALSE;
//				}
//			type t = UNPV(type, vec_fetch(v, 0));
//			string type_str = simple_type_name(type_name(t));
//
//			ind();
//			ps("{\n");
//			ind();
//			ps("struct Vector_f_s *");
//			expr_emit_local(prim);
//			ps("FS = ");
//
//			ps("RESIDENT_CORE(");
//			expr_emit(prim);
//			ps(", Vector);\n");
//			ind();
//			ps("PREPARE_READ_F(");
//			expr_emit(prim);
//			ps(", ");
//			expr_emit_local(prim);
//			ps("FS);\n");
//			ps("MARK_USED_F(");
//			expr_emit_local(prim);
//			ps("FS);\n");
//
//			// ind();
//			// expr_emit(prim);
//			// ps("FS = FIX_FIELDS(");
//			// expr_emit(prim);
//			// ps(", Vector);\n");
//
//   			ind();
//			string index_str = get_next_temp();
//			string limit_str = get_next_temp();
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps(" = ");
//			expr_emit_local(prim);
//			ps("FS->num_slots/4;\n");
//
//   			ind();
//   			ps("for (");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps(" = 0; ");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps(" < ");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps("; ");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps("++)");
//
//
//			ps(" {\n");
//			// ind();
//			// ps("DISCOVER(((vector_C)");
//			// expr_emit(prim);
//			// ps(")->items[");
//			// pstr(index_str);
//			// ps("], pval);\n");
//			ind();
//			emit_iter_prefix();
//			pstr(nm);
//			if (is_vtype(t)) ps(" = PV_as_prim(");
//			else ps(" = PV_as_obj(");
//			pstr(type_str);
//			ps(", ");
//			expr_emit_local(prim);
//			ps("FS->items[");
//			emit_iter_prefix();
//			pstr(index_str);
//			ps("]);\n");
//			
//			handle_body_1(body);
//			ind();
//			ps("}\n");
//			ind();
//			ps("}\n");
//			loop_end_label();
//			loop_exit();
//			return TRUE;
//			}
                case VECTOR_ELTS: {
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			loop_enter();
			ind();
			ps("{\n");
			ind();
			if (is_vtype(t)) {
				ps("VECTOR_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("VECTOR_GEN_PRIMITIVE_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), ");
                        	emit_iter_prefix();
				pstr(nm);
				ps(", ");
				pstr(simple_type_name(type_name(t)));
				ps(", (");
                        	emit_iter_prefix();
				pstr(cont);
				ps("));\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
				}
			else {
				ps("VECTOR_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("VECTOR_GEN_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), MT_1, ");
                        	emit_iter_prefix();
				pstr(cont);
				ps(");\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
                        	ind();
                        	emit_iter_prefix();
                        	pstr(nm);
                        	ps(" = PV_as_obj");
                        	ps("(");
                        	pstr(simple_type_name(type_name(t)));
                        	ps(", MT_1);\n");
				}
			handle_body_1(body);
			ind();
			ps("}\n");
			ind();
			ps("}\n");
			loop_end_label();
			loop_exit();
			return TRUE;
			}
		   else {
                        loop_enter();
                        ind();
                        ps("{\n");
                        ind();
                        if (is_vtype(t)) {
                                ps("VECTOR_PRIMITIVE_ELEMENTS_START(");
                                expr_emit(prim);
                                ps(", ");
                                emit_iter_prefix();
                                pstr(nm);
                                ps(", ");
                                pstr(simple_type_name(type_name(t)));
                                ps(");\n");
                                }
                        else {
                                ps("VECTOR_ELEMENTS_START(");
                                expr_emit(prim);
                                ps(", MT_1);\n");
                                ind();
                                emit_iter_prefix();
                                pstr(nm);
                                ps(" = PV_as_obj");
                                ps("(");
                                pstr(simple_type_name(type_name(t)));
                                ps(", MT_1);\n");
                                }
                        handle_body_1(body);
                        ind();
                        ps("VECTOR_ELEMENTS_END();\n");
                        ind();
                        ps("}\n");
                        loop_end_label();
                        loop_exit();
                        return TRUE;
                        }
                  }

                case VECTOR_INDX: {
                   loop_enter();
                   ind();
                   ps("{\n");
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			ind();
			ps("VECTOR_GEN_INDEXES_START(");
			expr_emit(prim);
			ps(", (");
                        emit_iter_prefix();
			pstr(state);
			ps("));\n");
			ind();
			ps("while (1) {\n");
			ind();
			ps("VECTOR_GEN_INDEXES_LOOP((");
                        emit_iter_prefix();
			pstr(state);
			ps("), ");
                        emit_iter_prefix();
			pstr(nm);
			ps(", ");
                        emit_iter_prefix();
			pstr(cont);
			ps(");\n");
			ind();
			ps("if (!");
			emit_iter_prefix();
			pstr(cont);
			ps(") break;\n");
			handle_body_1(body);
			ind();
			ps("}\n");
			}
		   else {
                        ind();
                        ps("VECTOR_INDEXES_START(");
                        expr_emit(prim);
                        ps(", ");
                        emit_iter_prefix();
                        pstr(nm);
                        ps(");\n");
                        handle_body_1(body);
                        ind();
                        ps("VECTOR_INDEXES_END();\n");
                        }
                   ind();
                   ps("}\n");
                   loop_end_label();
                   loop_exit();
                   return TRUE;
                   }

//		case VECTOR_INDX: {
//			loop_enter();
//			ind();
//			ps("{\n");
//
//			ind();
//			ps("struct Vector_f_s *");
//			expr_emit_local(prim);
//			ps("FS = ");
//			ps("RESIDENT_CORE(");
//			expr_emit(prim);
//			ps(", Vector);\n");
//			ind();
//			ps("PREPARE_READ_F(");
//			expr_emit(prim);
//			ps(", ");
//			expr_emit_local(prim);
//			ps("FS);\n");
//			ps("MARK_USED_F(");
//			expr_emit_local(prim);
//			ps("FS);\n");
//
//			// ind();
//			// expr_emit(prim);
//			// ps("FS = FIX_FIELDS(");
//			// expr_emit(prim);
//			// ps(", Vector);\n");
//
//   			ind();
//			string limit_str = get_next_temp();
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps(" = ");
//			expr_emit_local(prim);
//			ps("FS->num_slots/4;\n");
//
//   			ind();
//   			ps("for (");
//			emit_iter_prefix();
//			pstr(nm);
//			ps(" = 0; ");
//			emit_iter_prefix();
//			pstr(nm);
//			ps(" < ");
//			emit_iter_prefix();
//			pstr(limit_str);
//			ps("; ");
//			emit_iter_prefix();
//			pstr(nm);
//			ps("++)");
//
//			ps(" {\n");
//			handle_body_1(body);
//			ind();
//			ps("}\n");
//			ps("}\n");
//			loop_end_label();
//			loop_exit();
//			return TRUE;
//			}
                case SEQUENCE_INDX: {
                   loop_enter();
                   ind();
                   ps("{\n");
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			ind();
			ps("SEQUENCE_GEN_INDEXES_START(");
			expr_emit(prim);
			ps(", (");
                        emit_iter_prefix();
			pstr(state);
			ps("));\n");
			ind();
			ps("while (1) {\n");
			ind();
			ps("SEQUENCE_GEN_INDEXES_LOOP((");
                        emit_iter_prefix();
			pstr(state);
			ps("), ");
                        emit_iter_prefix();
			pstr(nm);
			ps(", ");
                        emit_iter_prefix();
			pstr(cont);
			ps(");\n");
			ind();
			ps("if (!");
			emit_iter_prefix();
			pstr(cont);
			ps(") break;\n");
			handle_body_1(body);
			ind();
			ps("}\n");
			}
		   else {
                        ind();
                        ps("SEQUENCE_INDEXES_START(");
                        expr_emit(prim);
                        ps(", ");
                        emit_iter_prefix();
                        pstr(nm);
                        ps(");\n");
                        handle_body_1(body);
                        ind();
                        ps("SEQUENCE_INDEXES_END();\n");
                        }
                   ind();
                   ps("}\n");
                   loop_end_label();
                   loop_exit();
                   return TRUE;
                   }
                case SEQUENCE_ELTS: {
                   loop_enter();
                   ind();
                   ps("{\n");
		   if (get_iter_active()) {
			string state = get_next_temp();
			string cont = get_next_temp();
			ind();
			if (is_vtype(t)) {
				ps("SEQUENCE_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("SEQUENCE_GEN_PRIMITIVE_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), ");
                        	emit_iter_prefix();
				pstr(nm);
				ps(", ");
				pstr(simple_type_name(type_name(t)));
				ps(", (");
                        	emit_iter_prefix();
				pstr(cont);
				ps("));\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
				}
			else {
				ps("SEQUENCE_GEN_ELEMENTS_START(");
				expr_emit(prim);
				ps(", (");
                        	emit_iter_prefix();
				pstr(state);
				ps("));\n");
				ind();
				ps("while (1) {\n");
				ind();
				ps("SEQUENCE_GEN_ELEMENTS_LOOP((");
                        	emit_iter_prefix();
				pstr(state);
				ps("), MT_1, ");
                        	emit_iter_prefix();
				pstr(cont);
				ps(");\n");
				ind();
				ps("if (!");
				emit_iter_prefix();
				pstr(cont);
				ps(") break;\n");
                        	ind();
                        	emit_iter_prefix();
                        	pstr(nm);
                        	ps(" = PV_as_obj");
                        	ps("(");
                        	pstr(simple_type_name(type_name(t)));
                        	ps(", MT_1);\n");
				}
			handle_body_1(body);
			ind();
			ps("}\n");
			}
		   else {
                        ind();
                        if (is_vtype(t)) {
                                ps("SEQUENCE_PRIMITIVE_ELEMENTS_START(");
                                expr_emit(prim);
                                ps(", ");
                                emit_iter_prefix();
                                pstr(nm);
                                ps(", ");
                                pstr(simple_type_name(type_name(t)));
                                ps(");\n");
                                }
                        else {
                                ps("SEQUENCE_ELEMENTS_START(");
                                expr_emit(prim);
                                ps(", MT_1);\n");
                                ind();
                                emit_iter_prefix();
                                pstr(nm);
                                ps(" = PV_as_obj");
                                ps("(");
                                pstr(simple_type_name(type_name(t)));
                                ps(", MT_1);\n");
                                }
                        handle_body_1(body);
                        ind();
                        ps("SEQUENCE_ELEMENTS_END();\n");
                        }
		ind();
		ps("}\n");
		loop_end_label();
		loop_exit();
		return TRUE;
                }
		}
	return FALSE;
	}

// Decide if this invoc is an inlinable iterator
// 	and, if so, indicate which one

bool opt_inv(Invoc *inv, iter_type *which)
{
Id *id;

	Expr *prim = opt_inv_get_prim(inv, &id);
	if (!prim) return FALSE;
	TypeInterface *ti = prim->get_type();
	type t = get_one_type(ti, prim->get_line());
	if (!t) return FALSE;
	const char *tname = string_charp(simple_type_name(type_name(t)));
	const char *mname = string_charp(id->get_id());
	if (!strcmp(tname, "int") && !strcmp(mname, "to")) {
		*which = INT_TO;
		return TRUE;
		}
	if (!strcmp(tname, "int") && !strcmp(mname, "to_by")) {
		*which = INT_TO_BY;
		return FALSE;
		}
	if (!strcmp(tname, "array") && !strcmp(mname, "elements")) {
		*which = ARRAY_ELTS;
		if (version == 0 || version == 1) return FALSE;
		else return TRUE;
		}
	if (!strcmp(tname, "array") && !strcmp(mname, "indexes")) {
		*which = ARRAY_INDX;
		return TRUE;
		}
	if (!strcmp(tname, "vector") && !strcmp(mname, "elements")) {
		*which = VECTOR_ELTS;
		return TRUE;
		}
	if (!strcmp(tname, "vector") && !strcmp(mname, "indexes")) {
		*which = VECTOR_INDX;
		return TRUE;
		}
	if (!strcmp(tname, "sequence") && !strcmp(mname, "elements")) {
		*which = SEQUENCE_ELTS;
		return TRUE;
		}
	if (!strcmp(tname, "sequence") && !strcmp(mname, "indexes")) {
		*which = SEQUENCE_INDX;
		return TRUE;
		}
	return FALSE;
	}

// Assuming that this invoc has a dot expr for its routineid, return
// 	the primary of the dot expr and the id of the dot expr.
// Returns a NULL pointer if this is not the case.

Expr *opt_inv_get_prim(Invoc *inv, Id **id)
{
Expr *ex;

	RoutineId *rid = inv->get_routineId();
	TypeInterface *ti = rid->get_type();
	method m = ti->get_method();
	if (!m) return NULL;
	switch (rid->tag()) {
		case RoutineId::SimpleRoutineIdT: {
			SimpleRoutineId *srid = (SimpleRoutineId *)rid;
			ex = srid->get_primary();
			break;
			}
		default: {
			return NULL;
			}
		}
	switch (ex->tag()) {
		case Expr::DotExprT: {
			DotExpr *de = (DotExpr *)ex;
			Expr *deid = de->get_id();
			switch (deid->tag()) {
				case Expr::IdExprT: {
					IdExpr *ide = (IdExpr *)deid;
		                        *id = ide->get_id();
                        		return de->get_primary();
					}
				}
			}
		}
	return NULL;
	}

bool opt_forstmt(ForStmt *forstmt)
{
iter_type which;

	Invoc *inv = forstmt->get_invoc();

	// Is this an inlineable iterator?
	// (Need to decide if it's inlineable *before* any code is emitted...)

	bool ok = opt_inv(inv, &which);
	if (!ok) return FALSE;

	// fish out the id

	ParseNodeList *ids = forstmt->get_ids();
	if (!ids || ids->length() != 1) {
			cmp_err("Inlining is incorrect.",
				inv->get_line());
			return FALSE;
			}
	Id *id = (Id *)(ids->front());

	Body *body = forstmt->get_body();

	return opt_for(which, id, inv, body);
	}
